module Calc where

import Parser
import Data.Char 
import Prelude hiding (catch)
import Control.Exception

data Expr   = Plus  Expr Expr
            | Times Expr Expr
            | Var    String
            | Num    Integer
            deriving (Show)

data Input  = Bind String Expr
            | Eval Expr
            deriving (Show)

-- Variable ::= Char+
pVar :: Parse Char String
pVar = some (spot isAlpha)

-- Number ::= Digit+
pNum :: Parse Char Integer
pNum = some (spot isDigit) `use` read

-- Factor ::= Variable | Number | '(' Expr ')'
pFactor :: Parse Char Expr
pFactor = pVar `use` Var
         `alt` pNum `use` Num 
         `alt` token '(' *> pExpr >* token ')'

-- Term ::= Factor '*' Term  | Factor
pTerm :: Parse Char Expr
pTerm = 
  pFactor >* token '*' >*> pTerm `use2` Times
  `alt` pFactor 

-- Expr ::= Term '+' Expr | Term
pExpr :: Parse Char Expr
pExpr = pTerm >* token '+' >*> pExpr `use2` Plus
      `alt` pTerm 

-- Auxiliary: spot a string
tokens :: String-> Parse Char ()
tokens str = foldr (\c p-> token c *> p) (suceed ()) str

-- Input ::= '!' Var '=' Expr | '$' Expr
pInput :: Parse Char Input
pInput = token '!' *> pVar >* token '=' >*> pExpr `use2` Bind
         `alt` 
         token '$' *> pExpr `use` Eval

-- Alternative Syntax: Input ::= 'let' Var '=' Expr | 'eval' Expr 
pInput2 :: Parse Char Input
pInput2 = tokens "let" *> pVar >* token '=' >*> pExpr `use2` Bind
         `alt` 
          tokens "eval" *> pExpr `use` Eval
  
-- Hauptfunktion
parseInput :: String-> Input
parseInput i = 
  case parse pInput (filter (not.isSpace) i) of 
    Right e ->  e
    Left err -> error err


-- Evaluation

-- We need a state to keep track of variables
type State= [(String, Integer)]

-- Evaluating an expression
eval :: State-> Expr-> Integer
eval s (Var i) = 
  case lookup i s of Nothing -> error $ "Undefined variable "++i
                     Just v -> v
eval s (Num i) = i
eval s (Plus e f)= eval s e+ eval s f
eval s (Times e f)= eval s e* eval s f

-- Parse and evaluate an input
run :: State-> String-> (State, String)
run s i = 
  case parseInput i of 
    Bind v e -> let w= eval s e 
                in ((v,w): s, v++ " set to "++ show w)
    Eval e   -> (s, show (eval s e))

-- Run a list of inputs, producing a final state and a list of outputs
run2 :: State-> [String]-> (State, [String])
run2 initial = 
  foldl (\(st, res) c-> let (st', res')= run st c in (st', res++ [res'])) 
        (initial, [])

-- Magic!
main :: IO ()
main = mainrun [] where
  mainrun s = do 
   i <- getLine
   (s', r) <- evaluate (run s i)
              `catch` (\(ErrorCall n)-> return (s, "*** "++ n))
   putStrLn r
   mainrun s'
              
